@@ -25,15 +25,22 @@ from ganeti import _autoconf
 
 # various versions
 CONFIG_VERSION = 3
-PROTOCOL_VERSION = 7
+PROTOCOL_VERSION = 13
 RELEASE_VERSION = _autoconf.PACKAGE_VERSION
 OS_API_VERSION = 5
 EXPORT_VERSION = 0
+RAPI_VERSION = 1
 
 
 # file paths
 DATA_DIR = _autoconf.LOCALSTATEDIR + "/lib/ganeti"
-BDEV_CACHE_DIR = _autoconf.LOCALSTATEDIR + "/run/ganeti"
+RUN_DIR = _autoconf.LOCALSTATEDIR + "/run"
+RUN_GANETI_DIR = RUN_DIR + "/ganeti"
+BDEV_CACHE_DIR = RUN_GANETI_DIR # TODO(2.0): move deeper
+DISK_LINKS_DIR = RUN_GANETI_DIR + "/instance-disks"
+# keep RUN_GANETI_DIR first here, to make sure all get created when the node
+# daemon is started (this takes care of RUN_DIR being tmpfs)
+SUB_RUN_DIRS = [ RUN_GANETI_DIR, BDEV_CACHE_DIR, DISK_LINKS_DIR ]
 LOCK_DIR = _autoconf.LOCALSTATEDIR + "/lock"
 CLUSTER_CONF_FILE = DATA_DIR + "/config.data"
 SSL_CERT_FILE = DATA_DIR + "/server.pem"
@@ -50,6 +57,8 @@ MASTER_SCRIPT = "ganeti-master"
 LOG_DIR = _autoconf.LOCALSTATEDIR + "/log/ganeti"
 LOG_OS_DIR = LOG_DIR + "/os"
 LOG_NODESERVER = LOG_DIR + "/node-daemon.log"
+LOG_RAPISERVER = LOG_DIR + "/rapi-daemon.log"
+LOG_RAPIACCESS = LOG_DIR + "/rapi-access.log"
 LOG_WATCHER = LOG_DIR + "/watcher.log"
 
 OS_SEARCH_PATH = _autoconf.OS_SEARCH_PATH
@@ -60,6 +69,9 @@ EXPORT_CONF_FILE = "config.ini"
 XEN_KERNEL = _autoconf.XEN_KERNEL
 XEN_INITRD = _autoconf.XEN_INITRD
 
+VALUE_DEFAULT = "default"
+VALUE_NONE = "none"
+
 # hooks-related constants
 HOOKS_BASE_DIR = _autoconf.SYSCONFDIR + "/ganeti/hooks"
 HOOKS_PHASE_PRE = "pre"
@@ -100,7 +112,7 @@ REPLACE_DISK_PRI = "replace_primary"
 REPLACE_DISK_SEC = "replace_secondary"
 REPLACE_DISK_ALL = "replace_all"
 
-# instance creation modem
+# instance creation modes
 INSTANCE_CREATE = "create"
 INSTANCE_IMPORT = "import"
 
@@ -114,6 +126,7 @@ INISECT_INS = "instance"
 
 # common exit codes
 EXIT_SUCCESS = 0
+EXIT_FAILURE = 1
 EXIT_NOTMASTER = 11
 EXIT_NODESETUP_ERROR = 12
 EXIT_CONFIRMATION = 13 # need user confirmation
@@ -122,8 +135,8 @@ EXIT_CONFIRMATION = 13 # need user confirmation
 TAG_CLUSTER = "cluster"
 TAG_NODE = "node"
 TAG_INSTANCE = "instance"
-MAX_TAG_LEN = 128
-MAX_TAGS_PER_OBJ = 4096
+MAX_TAG_LEN = 64
+MAX_TAGS_PER_OBJ = 48
 
 # others
 DEFAULT_BRIDGE = "xen-br0"
@@ -131,6 +144,15 @@ SYNC_SPEED = 30 * 1024
 LOCALHOST_IP_ADDRESS = "127.0.0.1"
 TCP_PING_TIMEOUT = 10
 GANETI_RUNAS = "root"
+BIND_ADDRESS_GLOBAL = "0.0.0.0"
+
+# migration rpc steps
+(DRBD_RECONF_RPC_INIT,
+ DRBD_RECONF_RPC_DISCONNECT,
+ DRBD_RECONF_RPC_RECONNECT,
+ DRBD_RECONF_RPC_SECONDARY,
+ DRBD_RECONF_RPC_WFSYNC,
+ ) = range(5)
 
 # valid os status
 OS_VALID_STATUS = "VALID"
@@ -147,3 +169,35 @@ SSH_HOST_RSA_PUB = SSH_HOST_RSA_PRIV + ".pub"
 INSTANCE_REBOOT_SOFT = "soft"
 INSTANCE_REBOOT_HARD = "hard"
 INSTANCE_REBOOT_FULL = "full"
+
+# Hypervisor constants
+HT_XEN_PVM30 = "xen-3.0"
+HT_FAKE = "fake"
+HT_XEN_HVM31 = "xen-hvm-3.1"
+HYPER_TYPES = frozenset([HT_XEN_PVM30, HT_FAKE, HT_XEN_HVM31])
+HTS_REQ_PORT = frozenset([HT_XEN_HVM31])
+
+HT_HVM_VNC_BASE_PORT = 5900
+HT_HVM_DEFAULT_BOOT_ORDER = 'dc'
+HT_HVM_DEFAULT_ACPI_MODE = '1'
+HT_HVM_DEFAULT_PAE_MODE = '1'
+VNC_PASSWORD_FILE = _autoconf.SYSCONFDIR + "/ganeti/vnc-cluster-password"
+VNC_DEFAULT_BIND_ADDRESS = '0.0.0.0'
+
+# Cluster Verify steps
+VERIFY_NPLUSONE_MEM = 'nplusone_mem'
+VERIFY_OPTIONAL_CHECKS = frozenset([VERIFY_NPLUSONE_MEM])
+
+# Allocator framework constants
+IALLOCATOR_DIR_IN = "in"
+IALLOCATOR_DIR_OUT = "out"
+IALLOCATOR_MODE_ALLOC = "allocate"
+IALLOCATOR_MODE_RELOC = "relocate"
+IALLOCATOR_SEARCH_PATH = _autoconf.IALLOCATOR_SEARCH_PATH
+IARUN_NOTFOUND = 1
+IARUN_FAILURE = 2
+IARUN_SUCCESS = 3
+
+# Remote API constants
+RAPI_ENABLE = _autoconf.RAPI_ENABLE
+RAPI_PORT = 5080
