@@ -25,18 +25,16 @@
 
 import time
 import os
+import re
 from cStringIO import StringIO
 
 from ganeti import utils
 from ganeti import logger
 from ganeti import ssconf
 from ganeti import constants
+from ganeti import errors
 from ganeti.errors import HypervisorError
 
-_HT_XEN30 = "xen-3.0"
-_HT_FAKE = "fake"
-
-VALID_HTYPES = (_HT_XEN30, _HT_FAKE)
 
 def GetHypervisor():
   """Return a Hypervisor instance.
@@ -46,10 +44,12 @@ def GetHypervisor():
 
   """
   ht_kind = ssconf.SimpleStore().GetHypervisorType()
-  if ht_kind == _HT_XEN30:
-    cls = XenHypervisor
-  elif ht_kind == _HT_FAKE:
+  if ht_kind == constants.HT_XEN_PVM30:
+    cls = XenPvmHypervisor
+  elif ht_kind == constants.HT_FAKE:
     cls = FakeHypervisor
+  elif ht_kind == constants.HT_XEN_HVM31:
+    cls = XenHvmHypervisor
   else:
     raise HypervisorError("Unknown hypervisor type '%s'" % ht_kind)
   return cls()
@@ -114,7 +114,7 @@ class BaseHypervisor(object):
     raise NotImplementedError
 
   @staticmethod
-  def GetShellCommandForConsole(instance_name):
+  def GetShellCommandForConsole(instance):
     """Return a command for connecting to the console of an instance.
 
     """
@@ -126,57 +126,41 @@ class BaseHypervisor(object):
     """
     raise NotImplementedError
 
+  def MigrateInstance(self, name, target, live):
+    """Migrate an instance.
+
+    Arguments:
+      - name: the name of the instance
+      - target: the target of the migration (usually will be IP and not name)
+      - live: whether to do live migration or not
+
+    Returns: none, errors will be signaled by exception.
+
+    """
+    raise NotImplementedError
+
 
 class XenHypervisor(BaseHypervisor):
-  """Xen hypervisor interface"""
+  """Xen generic hypervisor interface
+
+  This is the Xen base class used for both Xen PVM and HVM. It contains
+  all the functionality that is identical for both.
+
+  """
 
   @staticmethod
   def _WriteConfigFile(instance, block_devices, extra_args):
-    """Create a Xen 3.0 config file.
+    """Write the Xen config file for the instance.
 
     """
-    config = StringIO()
-    config.write("# this is autogenerated by Ganeti, please do not edit\n#\n")
-    config.write("kernel = '%s'\n" % constants.XEN_KERNEL)
-    if os.path.exists(constants.XEN_INITRD):
-      config.write("ramdisk = '%s'\n" % constants.XEN_INITRD)
-    config.write("memory = %d\n" % instance.memory)
-    config.write("vcpus = %d\n" % instance.vcpus)
-    config.write("name = '%s'\n" % instance.name)
-
-    vif_data = []
-    for nic in instance.nics:
-      nic_str = "mac=%s, bridge=%s" % (nic.mac, nic.bridge)
-      ip = getattr(nic, "ip", None)
-      if ip is not None:
-        nic_str += ", ip=%s" % ip
-      vif_data.append("'%s'" % nic_str)
-
-    config.write("vif = [%s]\n" % ",".join(vif_data))
-
-    disk_data = ["'phy:%s,%s,w'" % (rldev.dev_path, cfdev.iv_name)
-                 for cfdev, rldev in block_devices]
-    config.write("disk = [%s]\n" % ",".join(disk_data))
-
-    config.write("root = '/dev/sda ro'\n")
-    config.write("on_poweroff = 'destroy'\n")
-    config.write("on_reboot = 'restart'\n")
-    config.write("on_crash = 'restart'\n")
-    if extra_args:
-      config.write("extra = '%s'\n" % extra_args)
-    # just in case it exists
-    utils.RemoveFile("/etc/xen/auto/%s" % instance.name)
-    f = open("/etc/xen/%s" % instance.name, "w")
-    f.write(config.getvalue())
-    f.close()
-    return True
+    raise NotImplementedError
 
   @staticmethod
-  def _RemoveConfigFile(instance):
+  def _RemoveConfigFile(instance_name):
     """Remove the xen configuration file.
 
     """
-    utils.RemoveFile("/etc/xen/%s" % instance.name)
+    utils.RemoveFile("/etc/xen/%s" % instance_name)
 
   @staticmethod
   def _GetXMList(include_node):
@@ -200,12 +184,8 @@ class XenHypervisor(BaseHypervisor):
       raise HypervisorError("xm list failed, retries exceeded (%s): %s" %
                             (result.fail_reason, result.stderr))
 
-    # skip over the heading and the domain 0 line (optional)
-    if include_node:
-      to_skip = 1
-    else:
-      to_skip = 2
-    lines = result.stdout.splitlines()[to_skip:]
+    # skip over the heading
+    lines = result.stdout.splitlines()[1:]
     result = []
     for line in lines:
       # The format of lines is:
@@ -222,7 +202,11 @@ class XenHypervisor(BaseHypervisor):
       except ValueError, err:
         raise HypervisorError("Can't parse output of xm list,"
                               " line: %s, error: %s" % (line, err))
-      result.append(data)
+
+      # skip the Domain-0 (optional)
+      if include_node or data[0] != 'Domain-0':
+        result.append(data)
+
     return result
 
   def ListInstances(self):
@@ -270,7 +254,7 @@ class XenHypervisor(BaseHypervisor):
 
   def StopInstance(self, instance, force=False):
     """Stop an instance."""
-    self._RemoveConfigFile(instance)
+    self._RemoveConfigFile(instance.name)
     if force:
       command = ["xm", "destroy", instance.name]
     else:
@@ -317,6 +301,8 @@ class XenHypervisor(BaseHypervisor):
           result['memory_total'] = int(val)
         elif key == 'free_memory':
           result['memory_free'] = int(val)
+        elif key == 'nr_cpus':
+          result['cpu_total'] = int(val)
     dom0_info = self.GetInstanceInfo("Domain-0")
     if dom0_info is not None:
       result['memory_dom0'] = dom0_info[2]
@@ -324,11 +310,11 @@ class XenHypervisor(BaseHypervisor):
     return result
 
   @staticmethod
-  def GetShellCommandForConsole(instance_name):
+  def GetShellCommandForConsole(instance):
     """Return a command for connecting to the console of an instance.
 
     """
-    return "xm console %s" % instance_name
+    raise NotImplementedError
 
 
   def Verify(self):
@@ -340,6 +326,122 @@ class XenHypervisor(BaseHypervisor):
     if not utils.CheckDaemonAlive('/var/run/xend.pid', 'xend'):
       return "xend daemon is not running"
 
+  def MigrateInstance(self, instance, target, live):
+    """Migrate an instance to a target node.
+
+    Arguments:
+      - instance: the name of the instance
+      - target: the ip of the target node
+      - live: whether to do live migration or not
+
+    Returns: none, errors will be signaled by exception.
+
+    The migration will not be attempted if the instance is not
+    currently running.
+
+    """
+    if self.GetInstanceInfo(instance) is None:
+      raise errors.HypervisorError("Instance not running, cannot migrate")
+    args = ["xm", "migrate"]
+    if live:
+      args.append("-l")
+    args.extend([instance, target])
+    result = utils.RunCmd(args)
+    if result.failed:
+      raise errors.HypervisorError("Failed to migrate instance %s: %s" %
+                                   (instance, result.output))
+    # remove old xen file after migration succeeded
+    try:
+      self._RemoveConfigFile(instance)
+    except EnvironmentError, err:
+      logger.Error("Failure while removing instance config file: %s" %
+                   str(err))
+
+
+class XenPvmHypervisor(XenHypervisor):
+  """Xen PVM hypervisor interface"""
+
+  @staticmethod
+  def _WriteConfigFile(instance, block_devices, extra_args):
+    """Write the Xen config file for the instance.
+
+    """
+    config = StringIO()
+    config.write("# this is autogenerated by Ganeti, please do not edit\n#\n")
+
+    # kernel handling
+    if instance.kernel_path in (None, constants.VALUE_DEFAULT):
+      kpath = constants.XEN_KERNEL
+    else:
+      if not os.path.exists(instance.kernel_path):
+        raise errors.HypervisorError("The kernel %s for instance %s is"
+                                     " missing" % (instance.kernel_path,
+                                                   instance.name))
+      kpath = instance.kernel_path
+    config.write("kernel = '%s'\n" % kpath)
+
+    # initrd handling
+    if instance.initrd_path in (None, constants.VALUE_DEFAULT):
+      if os.path.exists(constants.XEN_INITRD):
+        initrd_path = constants.XEN_INITRD
+      else:
+        initrd_path = None
+    elif instance.initrd_path == constants.VALUE_NONE:
+      initrd_path = None
+    else:
+      if not os.path.exists(instance.initrd_path):
+        raise errors.HypervisorError("The initrd %s for instance %s is"
+                                     " missing" % (instance.initrd_path,
+                                                   instance.name))
+      initrd_path = instance.initrd_path
+
+    if initrd_path:
+      config.write("ramdisk = '%s'\n" % initrd_path)
+
+    # rest of the settings
+    config.write("memory = %d\n" % instance.memory)
+    config.write("vcpus = %d\n" % instance.vcpus)
+    config.write("name = '%s'\n" % instance.name)
+
+    vif_data = []
+    for nic in instance.nics:
+      nic_str = "mac=%s, bridge=%s" % (nic.mac, nic.bridge)
+      ip = getattr(nic, "ip", None)
+      if ip is not None:
+        nic_str += ", ip=%s" % ip
+      vif_data.append("'%s'" % nic_str)
+
+    config.write("vif = [%s]\n" % ",".join(vif_data))
+
+    disk_data = ["'phy:%s,%s,w'" % names for names in block_devices]
+    config.write("disk = [%s]\n" % ",".join(disk_data))
+
+    config.write("root = '/dev/sda ro'\n")
+    config.write("on_poweroff = 'destroy'\n")
+    config.write("on_reboot = 'restart'\n")
+    config.write("on_crash = 'restart'\n")
+    if extra_args:
+      config.write("extra = '%s'\n" % extra_args)
+    # just in case it exists
+    utils.RemoveFile("/etc/xen/auto/%s" % instance.name)
+    try:
+      f = open("/etc/xen/%s" % instance.name, "w")
+      try:
+        f.write(config.getvalue())
+      finally:
+        f.close()
+    except IOError, err:
+      raise errors.OpExecError("Cannot write Xen instance confile"
+                               " file /etc/xen/%s: %s" % (instance.name, err))
+    return True
+
+  @staticmethod
+  def GetShellCommandForConsole(instance):
+    """Return a command for connecting to the console of an instance.
+
+    """
+    return "xm console %s" % instance.name
+
 
 class FakeHypervisor(BaseHypervisor):
   """Fake hypervisor interface.
@@ -348,7 +450,7 @@ class FakeHypervisor(BaseHypervisor):
   a real virtualisation software installed.
 
   """
-  _ROOT_DIR = "/var/run/ganeti-fake-hypervisor"
+  _ROOT_DIR = constants.RUN_DIR + "/ganeti-fake-hypervisor"
 
   def __init__(self):
     BaseHypervisor.__init__(self)
@@ -495,12 +597,24 @@ class FakeHypervisor(BaseHypervisor):
           sum_free += int(val.split()[0])/1024
         elif key == 'Active':
           result['memory_dom0'] = int(val.split()[0])/1024
-
     result['memory_free'] = sum_free
+
+    cpu_total = 0
+    try:
+      fh = open("/proc/cpuinfo")
+      try:
+        cpu_total = len(re.findall("(?m)^processor\s*:\s*[0-9]+\s*$",
+                                   fh.read()))
+      finally:
+        fh.close()
+    except EnvironmentError, err:
+      raise HypervisorError("Failed to list node info: %s" % err)
+    result['cpu_total'] = cpu_total
+
     return result
 
   @staticmethod
-  def GetShellCommandForConsole(instance_name):
+  def GetShellCommandForConsole(instance):
     """Return a command for connecting to the console of an instance.
 
     """
@@ -515,3 +629,129 @@ class FakeHypervisor(BaseHypervisor):
     """
     if not os.path.exists(self._ROOT_DIR):
       return "The required directory '%s' does not exist." % self._ROOT_DIR
+
+
+class XenHvmHypervisor(XenHypervisor):
+  """Xen HVM hypervisor interface"""
+
+  @staticmethod
+  def _WriteConfigFile(instance, block_devices, extra_args):
+    """Create a Xen 3.1 HVM config file.
+
+    """
+    config = StringIO()
+    config.write("# this is autogenerated by Ganeti, please do not edit\n#\n")
+    config.write("kernel = '/usr/lib/xen/boot/hvmloader'\n")
+    config.write("builder = 'hvm'\n")
+    config.write("memory = %d\n" % instance.memory)
+    config.write("vcpus = %d\n" % instance.vcpus)
+    config.write("name = '%s'\n" % instance.name)
+    if instance.hvm_pae is None:   # use default value if not specified
+      config.write("pae = %s\n" % constants.HT_HVM_DEFAULT_PAE_MODE)
+    elif instance.hvm_pae:
+      config.write("pae = 1\n")
+    else:
+      config.write("pae = 0\n")
+    if instance.hvm_acpi is None:  # use default value if not specified
+      config.write("acpi = %s\n" % constants.HT_HVM_DEFAULT_ACPI_MODE)
+    elif instance.hvm_acpi:
+      config.write("acpi = 1\n")
+    else:
+      config.write("acpi = 0\n")
+    config.write("apic = 1\n")
+    arch = os.uname()[4]
+    if '64' in arch:
+      config.write("device_model = '/usr/lib64/xen/bin/qemu-dm'\n")
+    else:
+      config.write("device_model = '/usr/lib/xen/bin/qemu-dm'\n")
+    if instance.hvm_boot_order is None:
+      config.write("boot = '%s'\n" % constants.HT_HVM_DEFAULT_BOOT_ORDER)
+    else:
+      config.write("boot = '%s'\n" % instance.hvm_boot_order)
+    config.write("sdl = 0\n")
+    config.write("usb = 1\n");
+    config.write("usbdevice = 'tablet'\n");
+    config.write("vnc = 1\n")
+    config.write("vnclisten = '%s'\n" % instance.vnc_bind_address)
+
+    if instance.network_port > constants.HT_HVM_VNC_BASE_PORT:
+      display = instance.network_port - constants.HT_HVM_VNC_BASE_PORT
+      config.write("vncdisplay = %s\n" % display)
+      config.write("vncunused = 0\n")
+    else:
+      config.write("# vncdisplay = 1\n")
+      config.write("vncunused = 1\n")
+
+    try:
+      password_file = open(constants.VNC_PASSWORD_FILE, "r")
+      try:
+        password = password_file.readline()
+      finally:
+        password_file.close()
+    except IOError:
+      raise errors.OpExecError("failed to open VNC password file %s " %
+                               constants.VNC_PASSWORD_FILE)
+
+    config.write("vncpasswd = '%s'\n" % password.rstrip())
+
+    config.write("serial = 'pty'\n")
+    config.write("localtime = 1\n")
+
+    vif_data = []
+    for nic in instance.nics:
+      nic_str = "mac=%s, bridge=%s, type=ioemu" % (nic.mac, nic.bridge)
+      ip = getattr(nic, "ip", None)
+      if ip is not None:
+        nic_str += ", ip=%s" % ip
+      vif_data.append("'%s'" % nic_str)
+
+    config.write("vif = [%s]\n" % ",".join(vif_data))
+
+    # TODO(2.0): This code changes the block device name, seen by the instance,
+    # from what Ganeti believes it should be. Different hypervisors may have
+    # different requirements, so we should probably review the design of
+    # storing it altogether, for the next major version.
+    disk_data = ["'phy:%s,%s,w'" %
+                 (dev_path, iv_name.replace("sd", "ioemu:hd"))
+                 for dev_path, iv_name in block_devices]
+
+    if instance.hvm_cdrom_image_path is None:
+      config.write("disk = [%s]\n" % (",".join(disk_data)))
+    else:
+      iso = "'file:%s,hdc:cdrom,r'" % (instance.hvm_cdrom_image_path)
+      config.write("disk = [%s, %s]\n" % (",".join(disk_data), iso))
+
+    config.write("on_poweroff = 'destroy'\n")
+    config.write("on_reboot = 'restart'\n")
+    config.write("on_crash = 'restart'\n")
+    if extra_args:
+      config.write("extra = '%s'\n" % extra_args)
+    # just in case it exists
+    utils.RemoveFile("/etc/xen/auto/%s" % instance.name)
+    try:
+      f = open("/etc/xen/%s" % instance.name, "w")
+      try:
+        f.write(config.getvalue())
+      finally:
+        f.close()
+    except IOError, err:
+      raise errors.OpExecError("Cannot write Xen instance confile"
+                               " file /etc/xen/%s: %s" % (instance.name, err))
+    return True
+
+  @staticmethod
+  def GetShellCommandForConsole(instance):
+    """Return a command for connecting to the console of an instance.
+
+    """
+    if instance.network_port is None:
+      raise errors.OpExecError("no console port defined for %s"
+                               % instance.name)
+    elif instance.vnc_bind_address == constants.BIND_ADDRESS_GLOBAL:
+      raise errors.OpExecError("no PTY console, connect to %s:%s via VNC"
+                               % (instance.primary_node,
+                                  instance.network_port))
+    else:
+      raise errors.OpExecError("no PTY console, connect to %s:%s via VNC"
+                               % (instance.vnc_bind_address,
+                                  instance.network_port))
