@@ -27,7 +27,6 @@ pass to and from external parties.
 """
 
 
-import simplejson
 import ConfigParser
 import re
 from cStringIO import StringIO
@@ -40,14 +39,6 @@ __all__ = ["ConfigObject", "ConfigData", "NIC", "Disk", "Instance",
            "OS", "Node", "Cluster"]
 
 
-# Check whether the simplejson module supports indentation
-_JSON_INDENT = 2
-try:
-  simplejson.dumps(1, indent=_JSON_INDENT)
-except TypeError:
-  _JSON_INDENT = None
-
-
 class ConfigObject(object):
   """A generic config object.
 
@@ -90,34 +81,6 @@ class ConfigObject(object):
       if name in self.__slots__:
         setattr(self, name, state[name])
 
-  def Dump(self, fobj):
-    """Dump to a file object.
-
-    """
-    data = self.ToDict()
-    if _JSON_INDENT is None:
-      simplejson.dump(data, fobj)
-    else:
-      simplejson.dump(data, fobj, indent=_JSON_INDENT)
-
-  @classmethod
-  def Load(cls, fobj):
-    """Load data from the given stream.
-
-    """
-    return cls.FromDict(simplejson.load(fobj))
-
-  def Dumps(self):
-    """Dump and return the string representation."""
-    buf = StringIO()
-    self.Dump(buf)
-    return buf.getvalue()
-
-  @classmethod
-  def Loads(cls, data):
-    """Load data from a string."""
-    return cls.Load(StringIO(data))
-
   def ToDict(self):
     """Convert to a dict holding only standard python types.
 
@@ -201,15 +164,21 @@ class TaggableObject(ConfigObject):
   __slots__ = ConfigObject.__slots__ + ["tags"]
 
   @staticmethod
-  def ValidateTag(tag):
+  def ValidateTag(tag, removal=False):
     """Check if a tag is valid.
 
     If the tag is invalid, an errors.TagError will be raised. The
     function has no return value.
 
+    Args:
+      tag: Tag value
+      removal: Validating tag for removal?
+
     """
     if not isinstance(tag, basestring):
       raise errors.TagError("Invalid tag type (not a string)")
+    if removal:
+      return
     if len(tag) > constants.MAX_TAG_LEN:
       raise errors.TagError("Tag too long (>%d characters)" %
                             constants.MAX_TAG_LEN)
@@ -241,7 +210,7 @@ class TaggableObject(ConfigObject):
     """Remove a tag.
 
     """
-    self.ValidateTag(tag)
+    self.ValidateTag(tag, removal=True)
     tags = self.GetTags()
     try:
       tags.remove(tag)
@@ -411,6 +380,68 @@ class Disk(ConfigObject):
             # be different)
     return result
 
+  def RecordGrow(self, amount):
+    """Update the size of this disk after growth.
+
+    This method recurses over the disks's children and updates their
+    size correspondigly. The method needs to be kept in sync with the
+    actual algorithms from bdev.
+
+    """
+    if self.dev_type == constants.LD_LV:
+      self.size += amount
+    elif self.dev_type == constants.LD_DRBD8:
+      if self.children:
+        self.children[0].RecordGrow(amount)
+      self.size += amount
+    else:
+      raise errors.ProgrammerError("Disk.RecordGrow called for unsupported"
+                                   " disk type %s" % self.dev_type)
+
+  def SetPhysicalID(self, target_node, nodes_ip):
+    """Convert the logical ID to the physical ID.
+
+    This is used only for drbd, which needs ip/port configuration.
+
+    The routine descends down and updates its children also, because
+    this helps when the only the top device is passed to the remote
+    node.
+
+    Arguments:
+      - target_node: the node we wish to configure for
+      - nodes_ip: a mapping of node name to ip
+
+    The target_node must exist in in nodes_ip, and must be one of the
+    nodes in the logical ID for each of the DRBD devices encountered
+    in the disk tree.
+
+    """
+    if self.children:
+      for child in self.children:
+        child.SetPhysicalID(target_node, nodes_ip)
+
+    if self.logical_id is None and self.physical_id is not None:
+      return
+    if self.dev_type in constants.LDS_DRBD:
+      pnode, snode, port = self.logical_id
+      if target_node not in (pnode, snode):
+        raise errors.ConfigurationError("DRBD device not knowing node %s" %
+                                        target_node)
+      pnode_ip = nodes_ip.get(pnode, None)
+      snode_ip = nodes_ip.get(snode, None)
+      if pnode_ip is None or snode_ip is None:
+        raise errors.ConfigurationError("Can't find primary or secondary node"
+                                        " for %s" % str(self))
+      if pnode == target_node:
+        self.physical_id = (pnode_ip, port,
+                            snode_ip, port)
+      else: # it must be secondary, we tested above
+        self.physical_id = (snode_ip, port,
+                            pnode_ip, port)
+    else:
+      self.physical_id = self.logical_id
+    return
+
   def ToDict(self):
     """Disk-specific conversion to standard python types.
 
@@ -490,6 +521,14 @@ class Instance(TaggableObject):
     "nics",
     "disks",
     "disk_template",
+    "network_port",
+    "kernel_path",
+    "initrd_path",
+    "hvm_boot_order",
+    "hvm_acpi",
+    "hvm_pae",
+    "hvm_cdrom_image_path",
+    "vnc_bind_address",
     ]
 
   def _ComputeSecondaryNodes(self):
